/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#include "eb.h"
#include "error.h"
#include "font.h"

#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif

/*
 * Whence parameter for lseek().
 */
#ifndef SEEK_SET
#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#endif

/*
 * The maximum length of path name.
 */
#ifndef PATH_MAX
#ifdef MAXPATHLEN
#define PATH_MAX	MAXPATHLEN
#else /* not MAXPATHLEN */
#define PATH_MAX	1024
#endif /* not MAXPATHLEN */
#endif /* not PATH_MAX */


/*
 * Look up the height of the current font of the current subbook in
 * `book'.
 */
EB_Font_Code
eb_font(book)
    EB_Book *book;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Look up the height of the current font.
     */
    if (book->subcurrent->narwcurrent != NULL)
	return book->subcurrent->narwcurrent->height;

    if (book->subcurrent->widecurrent != NULL)
	return book->subcurrent->widecurrent->height;

    eb_error = EB_ERR_NO_CUR_FONT;
    return -1;
}


/*
 * Set the font with `height' as the current font of the current
 * subbook in `book'.
 */
int
eb_set_font(book, height)
    EB_Book *book;
    EB_Font_Code height;
{
    char filename[PATH_MAX + 1];
    EB_Subbook *sub = book->subcurrent;
    EB_Font *fnt;
    int narrowwidth, widewidth;
    int i;

    /*
     * Current subbook must have been set.
     */
    if (sub == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	goto failed;
    }

    /*
     * Get the width of the new fonts.
     */
    narrowwidth = eb_narrow_font_width2(height);
    widewidth = eb_wide_font_width2(height);
    if (narrowwidth < 0) {
	eb_error = EB_ERR_NO_SUCH_FONT;
	goto failed;
    }

    /*
     * If the current font is the font with `height', return immediately.
     * Otherwise close the current font and continue.
     */
    if (sub->narwcurrent != NULL) {
	if (sub->narwcurrent->height == height)
	    return 0;
	if (book->disccode == EB_DISC_EPWING)
	    close(sub->narwcurrent->fontfile);
	sub->narwcurrent = NULL;
    }
    if (sub->widecurrent != NULL) {
	if (sub->widecurrent->height == height)
	    return 0;
	if (book->disccode == EB_DISC_EPWING)
	    close(sub->widecurrent->fontfile);
	sub->widecurrent = NULL;
    }

    /*
     * Scan the font table in the book.
     */
    for (i = 0, fnt = sub->fonts; i < sub->fontcount; i++, fnt++) {
	if (fnt->height == height && fnt->width == narrowwidth) {
	    sub->narwcurrent = fnt;
	    break;
	}
    }
    for (i = 0, fnt = sub->fonts; i < sub->fontcount; i++, fnt++) {
	if (fnt->height == height && fnt->width == widewidth) {
	    sub->widecurrent = fnt;
	    break;
	}
    }

    if (sub->narwcurrent == NULL && sub->widecurrent == NULL) {
	eb_error = EB_ERR_NO_SUCH_FONT;
	goto failed;
    }

    /*
     * Set file discriptors for fonts.
     * If the book is EBWING, open font files.
     * (In EB books, font data are stored in the `START' file.)
     */
    if (sub->narwcurrent != NULL) {
	if (book->disccode == EB_DISC_EB)
	    sub->narwcurrent->fontfile = sub->startfile;
	else {
	    sprintf(filename, "%s/%s/%s/%s", book->path, sub->directory,
		EB_DIRNAME_GAIJI, sub->narwcurrent->filename);
	    eb_fix_filename(book, filename);
	    sub->narwcurrent->fontfile = open(filename, O_RDONLY);
	    if (sub->narwcurrent->fontfile < 0) {
		book->subcurrent->narwcurrent = NULL;
		eb_error = EB_ERR_FAIL_OPEN_FONT;
		goto failed;
	    }
	}
    }
    if (sub->widecurrent != NULL) {
	if (book->disccode == EB_DISC_EB)
	    sub->widecurrent->fontfile = sub->startfile;
	else {
	    sprintf(filename, "%s/%s/%s/%s", book->path, sub->directory,
		EB_DIRNAME_GAIJI, sub->widecurrent->filename);
	    eb_fix_filename(book, filename);
	    sub->widecurrent->fontfile = open(filename, O_RDONLY);
	    if (sub->widecurrent->fontfile < 0) {
		book->subcurrent->widecurrent = NULL;
		eb_error = EB_ERR_FAIL_OPEN_FONT;
		goto failed;
	    }
	}
    }

    return 0;

    /*
     * An error occurs...
     */
  failed:
    eb_unset_font(book);
    return -1;
}


/*
 * Unset the font in the current subbook in `book'.
 */
void
eb_unset_font(book)
    EB_Book *book;
{
    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL)
	return;

    /*
     * Close font files if the book is EPWING and font files are
     * opened.
     */
    if (book->disccode == EB_DISC_EPWING) {
	if (book->subcurrent->narwcurrent != NULL)
	    close(book->subcurrent->narwcurrent->fontfile);
	if (book->subcurrent->widecurrent != NULL)
	    close(book->subcurrent->widecurrent->fontfile);
    }

    book->subcurrent->narwcurrent = NULL;
    book->subcurrent->widecurrent = NULL;
}


/*
 * Get the number of fonts in the current subbook in `book'.
 */
int
eb_font_count(book)
    EB_Book *book;
{
    EB_Font *fnt1, *fnt2;
    int i, j;
    int count = 0;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return 0;
    }

    /*
     * Scan the font table in the book.
     */
    for (i = 0, fnt1 = book->subcurrent->fonts;
	 i < book->subcurrent->fontcount; i++, fnt1++) {
	for (j = 0, fnt2 = book->subcurrent->fonts; j < i; j++, fnt2++) {
	    if (fnt1->height == fnt2->height)
		break;
	}
	if (i <= j)
	    count++;
    }

    return count;
}


/*
 * Make a list of fonts in the current subbook in `book'.
 */
int
eb_font_list(book, list)
    EB_Book *book;
    EB_Font_Code *list;
{
    EB_Font_Code *lp = list;
    EB_Font *fnt1, *fnt2;
    int i, j;
    int count = 0;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return -1;
    }

    /*
     * Scan the font table in the book.
     */
    for (i = 0, fnt1 = book->subcurrent->fonts;
	 i < book->subcurrent->fontcount; i++, fnt1++) {
	for (j = 0, fnt2 = book->subcurrent->fonts; j < i; j++, fnt2++) {
	    if (fnt1->height == fnt2->height)
		break;
	}
	if (i <= j) {
	    *lp++ = fnt1->height;
	    count++;
	}
    }

    return count;
}


/*
 * Test whether the current subbook in `book' has a font with
 * `height' or not.
 */
int
eb_have_font(book, height)
    EB_Book *book;
    EB_Font_Code height;
{
    EB_Font *fnt;
    int i;

    /*
     * Current subbook must have been set.
     */
    if (book->subcurrent == NULL) {
	eb_error = EB_ERR_NO_CUR_SUB;
	return 0;
    }

    /*
     * Scan font entries.
     */
    for (i = 0, fnt = book->subcurrent->fonts;
	 i < book->subcurrent->fontcount; i++, fnt++) {
	if (fnt->height == height)
	    return 1;
    }

    eb_error = EB_ERR_NO_SUCH_FONT;
    return 0;
}


