/*
 * Copyright (c) 1997  Motoyuki Kasahara
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <syslog.h>

#if defined(STDC_HEADERS) || defined(HAVE_STRING_H)
#include <string.h>
#if !defined(STDC_HEADERS) && defined(HAVE_MEMORY_H)
#include <memory.h>
#endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#else /* not STDC_HEADERS and not HAVE_STRING_H */
#include <strings.h>
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifndef HAVE_STRCHR
#define strchr index
#define strrchr rindex
#endif /* HAVE_STRCHR */

#include <eb/eb.h>
#include <eb/appendix.h>

#include "confutil.h"
#include "filename.h"
#include "hostname.h"
#include "ndtpd.h"
#include "permission.h"
#include "readconf.h"

#ifdef USE_FAKELOG
#include "fakelog.h"
#endif


/*
 * Unexported functions.
 */
#ifdef __STDC__

static int conf_begin(const char *, const char *, const char *,
    const char *, int);
static int conf_end(const char *, const char *, const char *,
    const char *, int);
static int conf_server_name(const char *, const char *, const char *,
    const char *, int);
static int conf_ndtp_port(const char *, const char *, const char *,
    const char *, int);
static int conf_user(const char *, const char *, const char *,
    const char *, int);
static int conf_group(const char *, const char *, const char *,
    const char *, int);
static int conf_max_clients(const char *, const char *, const char *,
    const char *, int);
static int conf_hosts(const char *, const char *, const char *,
    const char *, int);
static int conf_timeout(const char *, const char *, const char *,
    const char *, int);
static int conf_ident_hosts(const char *, const char *, const char *,
    const char *, int);
static int conf_ident_timeout(const char *, const char *, const char *,
    const char *, int);
static int conf_work_path(const char *, const char *, const char *,
    const char *, int);
static int conf_max_hits(const char *, const char *, const char *,
    const char *, int);
static int conf_syslog_facility(const char *, const char *, const char *,
    const char *, int);
static int conf_begin_book(const char *, const char *, const char *,
    const char *, int);
static int conf_end_book(const char *, const char *, const char *,
    const char *, int);
static int conf_book_name(const char *, const char *, const char *,
    const char *, int);
static int conf_book_title(const char *, const char *, const char *,
    const char *, int);
static int conf_book_path(const char *, const char *, const char *,
    const char *, int);
static int conf_appendix_path(const char *, const char *, const char *,
    const char *, int);
static int conf_book_max_clients(const char *, const char *, const char *,
    const char *, int);
static int conf_book_hosts(const char *, const char *, const char *,
    const char *, int);
static int conf_book_alias_eiwa(const char *, const char *, const char *,
    const char *, int);
static int conf_book_alias_waei(const char *, const char *, const char *,
    const char *, int);
static int conf_book_alias_kojien(const char *, const char *, const char *,
    const char *, int);

#else /* not __STDC__ */

static int conf_begin();
static int conf_end();
static int conf_server_name();
static int conf_ndtp_port();
static int conf_user();
static int conf_group();
static int conf_max_clients();
static int conf_hosts();
static int conf_timeout();
static int conf_ident_hosts();
static int conf_ident_timeout();
static int conf_work_path();
static int conf_max_hits();
static int conf_syslog_facility();
static int conf_begin_book();
static int conf_end_book();
static int conf_book_name();
static int conf_book_title();
static int conf_book_path();
static int conf_appendix_path();
static int conf_book_max_clients();
static int conf_book_hosts();
static int conf_book_alias_eiwa();
static int conf_book_alias_waei();
static int conf_book_alias_kojien();

#endif /* not __STDC__ */


/*
 * Syntax of the configuration file.
 */
Configuration configuration_table[] = {
    {"",     "begin",           conf_begin,             0, 0, 0},
    {"",     "end",             conf_end,               0, 0, 0},
    {"",     "server-name",     conf_server_name,       0, 0, 0},
    {"",     "ndtp-port",       conf_ndtp_port,         0, 0, 0},
    {"",     "user",            conf_user,              0, 0, 0},
    {"",     "group",           conf_group,             0, 0, 0},
    {"",     "max-clients",     conf_max_clients,       0, 0, 0},
    {"",     "hosts",           conf_hosts,             0, 1, 0},
    {"",     "timeout",         conf_timeout,           0, 0, 0},
    {"",     "ident-hosts",     conf_ident_hosts,       0, 1, 0},
    {"",     "ident-timeout",   conf_ident_timeout,     0, 0, 0},
    {"",     "work-path",       conf_work_path,         0, 0, 0},
    {"",     "max-hits",        conf_max_hits,          0, 0, 0},
    {"",     "syslog-facility", conf_syslog_facility,   0, 0, 0},
    {"book", "begin",           conf_begin_book,        0, 0, 0},
    {"book", "end",             conf_end_book,          0, 0, 0},
    {"book", "name",            conf_book_name,         1, 0, 0},
    {"book", "title",           conf_book_title,        1, 0, 0},
    {"book", "path",            conf_book_path,         1, 0, 0},
    {"book", "appendix-path",   conf_appendix_path,     0, 0, 0},
    {"book", "max-clients",     conf_book_max_clients,  0, 0, 0},
    {"book", "hosts",           conf_book_hosts,        0, 1, 0},
    {"book", "alias-eiwa",      conf_book_alias_eiwa,   0, 0, 0},
    {"book", "alias-waei",      conf_book_alias_waei,   0, 0, 0},
    {"book", "alias-kojien",    conf_book_alias_kojien, 0, 0, 0},
    {"",     NULL,            NULL,                     0, 0, 0}
};


/*
 * The hook is called before reading a configuration file.
 */
static int
conf_begin(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char server_address[MAXLEN_HOSTNAME + 1];

    /*
     * Reset the book list.
     */
    clear_book_registry();

    /*
     * Reset the host permission list.
     */
    clear_permission(&permissions);

    /*
     * Reset the identification list.
     */
    clear_permission(&identifications);

    /*
     * Initialize global variables.
     */
    server_name[0] = '\0';
    user_id = 0;
    group_id = 0;
    strcpy(work_path, DEFAULT_WORK_PATH);
    max_clients = DEFAULT_MAX_CLIENTS;
    idle_timeout = DEFAULT_IDLE_TIMEOUT;
    identification_timeout = DEFAULT_IDENT_TIMEOUT;
    max_hits = DEFAULT_MAXHITS;
    listening_port = -1;
    alias_eiwa[0] = '\0';
    alias_waei[0] = '\0';
    alias_kojien[0] = '\0';

    /*
     * Get a host name of the server.
     */
    if (identify_my_host(server_name, server_address) < 0)
	syslog(LOG_ERR, "warning, cannot get a server hostname");

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s: debug: beginning of the configuration",
	filename);

    return 0;
}


/*
 * The hook is called just after reading a configuration file.
 */
static int
conf_end(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    static const char *pid_basename = PID_BASENAME;
    size_t worklen;

    /*
     * Terminate the book list `book_registry'.
     */
    terminate_book_registry();

    /*
     * Check for the server name.  It must contain a domain name.
     */
    if (strchr(server_name, '.') == NULL) {
	syslog(LOG_ERR, "the server name without domain name: %s",
	    server_name);
	return -1;
    }

    /*
     * Check for the server port number.
     */
    if (listening_port < 0
	&& parse_tcp_port(DEFAULT_PORT, &listening_port) < 0) {
	syslog(LOG_ERR, "unknown service name: %s", DEFAULT_PORT);
	return -1;
    }

    /*
     * Convert `pid_file' from relative path to absolute path.
     */
    worklen = strlen(work_path);
    if (MAXLEN_FILENAME < worklen + 1 + strlen(pid_basename)) {
	*(pid_filename + MAXLEN_FILENAME) = '\0';
	syslog(LOG_ERR, "%s: too long filename: %s...", filename,
	    work_path);
	return -1;
    }
    sprintf(pid_filename, "%s/%s", work_path, pid_basename);
    if (canonicalize_filename(pid_filename) < 0)
	return -1;
    syslog(LOG_DEBUG, "%s: debug: pid-filename is expaned: %s",
	filename, pid_filename);

    /*
     * debugging information.
     */
    syslog(LOG_DEBUG, "%s: debug: end of the configuration", filename);
    return 0;
}

    
/*
 * `server-name' directive.
 */
static int
conf_server_name(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    /*
     * Check for the length of the server name.
     */
    if (MAXLEN_HOSTNAME < strlen(arg)) {
	strncpy(server_name, arg, MAXLEN_HOSTNAME);
	*(server_name + MAXLEN_HOSTNAME) = '\0'; 
	syslog(LOG_ERR, "%s:%d: too long hostname: %s...", filename,
	    lineno, server_name);
	return -1;
    }

    /*
     * Set the server name.
     */
    strcpy(server_name, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set server-name: %s",
	filename, lineno, server_name);

    return 0;
}


/*
 * `ndtp-port' directive.
 */
static int
conf_ndtp_port(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the port number.
     */
    if (parse_tcp_port(arg, &listening_port) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: unknown service: %s", filename, lineno,
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set ndtp-port: %d", filename,
	lineno, listening_port);

    return 0;
}


/*
 * `user' directive.
 */
static int
conf_user(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the user ID.
     */
    if (parse_user(arg, &user_id) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: unknown user: %s", filename, lineno, tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set user: %d", filename, lineno,
	(int)user_id);

    return 0;
}


/*
 * `group' directive.
 */
static int
conf_group(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the group ID.
     */
    if (parse_group(arg, &group_id) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: unknown group: %s", filename, lineno, tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set group: %d", filename, lineno,
	(int)group_id);

    return 0;
}


/*
 * `max-clients' directive.
 */
static int
conf_max_clients(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the max-clients.
     */
    if (parse_integer(arg, &max_clients) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: not an integer: %s", filename, lineno, 
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set max-clients: %d", filename,
	lineno, max_clients);

    return 0;
}


/*
 * `hosts' directive.
 */
static int
conf_hosts(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Add `arg' to the host permission list.
     */
    if (add_permission(&permissions, arg) < 0) {
	syslog(LOG_ERR, "%s:%d: memory exhausted", filename, lineno);
	return -1;
    }

    /*
     * Debugging information.
     */
    strncpy(tmpbuf, arg, MAXLEN_STRING);
    *(tmpbuf + MAXLEN_STRING) = '\0';
    syslog(LOG_DEBUG, "%s:%d: debug: add hosts: %s", filename, lineno,
	tmpbuf);
    return 0;
}


/*
 * `timeout' directive.
 */
static int
conf_timeout(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the timeout.
     */
    if (parse_integer(arg, &idle_timeout) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: not an integer: %s", filename, lineno,
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set timeout: %d", filename,
	lineno, idle_timeout);

    return 0;
}


/*
 * `ident-hosts' directive.
 */
static int
conf_ident_hosts(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Add `arg' to the ideintification list.
     */
    if (add_permission(&identifications, arg) < 0) {
	syslog(LOG_ERR, "%s:%d: memory exhausted", filename, lineno);
	return -1;
    }

    /*
     * Debugging information.
     */
    strncpy(tmpbuf, arg, MAXLEN_STRING);
    *(tmpbuf + MAXLEN_STRING) = '\0';
    syslog(LOG_DEBUG, "%s:%d: debug: add ident: %s", filename, lineno,
	tmpbuf);

    return 0;
}


/*
 * `ident-timeout' directive.
 */
static int
conf_ident_timeout(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the ident timeout.
     */
    if (parse_integer(arg, &identification_timeout) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: not an integer: %s", filename, lineno,
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set ident-timeout: %d", filename,
	lineno, identification_timeout);

    return 0;
}


/*
 * `work-path' directive.
 */
static int
conf_work_path(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    /*
     * Check for the length of the path.
     */
    if (MAXLEN_FILENAME < strlen(arg)) {
	strncpy(work_path, arg, MAXLEN_FILENAME);
	*(work_path + MAXLEN_FILENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long path: %s...", filename, lineno,
	    work_path);
	return -1;
    }

    /*
     * The filename must be an absolute path.
     */
    if (*arg != '/') {
	syslog(LOG_ERR, "%s:%d: not an absolute path: %s", filename, lineno,
	    arg);
	return -1;
    }

    /*
     * Set the working directory path.
     */
    strcpy(work_path, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set work-path: %s", filename,
	lineno, work_path);

    return 0;
}


/*
 * `max-hits' directive.
 */
static int
conf_max_hits(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Parse and set the max hits.
     */
    if (parse_integer(arg, &max_hits) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: not an integer: %s", filename, lineno,
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set max-hits: %d", filename,
	lineno, max_hits);

    return 0;
}


/*
 * `syslog-facility' directive.
 */
static int
conf_syslog_facility(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    /*
     * Parse and set a syslog facility.
     */
    if (parse_syslog_facility(arg, &syslog_facility) < 0) {
	syslog(LOG_ERR, "%s:%d: unknown syslog facility: %s", filename,
	    lineno, arg);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set syslog-facility: %s", filename,
	lineno, arg);

    return 0;
}


/*
 * `begin book' sequence.
 */
static int
conf_begin_book(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    /*
     * Create a new book node.
     */
    if (add_book() < 0) {
	syslog(LOG_ERR, "%s:%d: memory exhausted", filename, lineno);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: beginning of the book definition",
	filename, lineno);

    return 0;
}


/*
 * `end book' sequence.
 */
static int
conf_end_book(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_BOOKNAME + 1 + EB_MAXLEN_BASENAME + 1];

    /*
     * Convert aliases to "book-name/subbook-name" style.
     */
    if (alias_eiwa[0] != '\0' && strchr(alias_eiwa, '/') == NULL) {
	sprintf(tmpbuf, "%s/%s", current_book->name, alias_eiwa);
	strcpy(alias_eiwa, tmpbuf);
    }
    if (alias_waei[0] != '\0' && strchr(alias_waei, '/') == NULL) {
	sprintf(tmpbuf, "%s/%s", current_book->name, alias_waei);
	strcpy(alias_waei, tmpbuf);
    }
    if (alias_kojien[0] != '\0' && strchr(alias_kojien, '/') == NULL) {
	sprintf(tmpbuf, "%s/%s", current_book->name, alias_kojien);
	strcpy(alias_kojien, tmpbuf);
    }

    /*
     * If failed to bind the book, delete it from a registry.
     */
    if (!eb_is_bound(&current_book->book))
	delete_book();

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: end of the book definition",
	filename, lineno);

    return 0;
}


/*
 * `name' directive (in book structure).
 */
static int
conf_book_name(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    const char *p;

    /*
     * Check for the length of the book name.
     */
    if (MAXLEN_BOOKNAME < strlen(arg)) {
	strncpy(current_book->name, arg, MAXLEN_BOOKNAME);
	*(current_book->name + MAXLEN_BOOKNAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long book name: %s...", filename, lineno,
	    current_book->name);
	return -1;
    }

    /*
     * The tagname must consist of upper letters, digit, '_', and '-'.
     */
    for (p = arg; *p != '\0'; p++) {
	if (!isupper(*p) && !isdigit(*p) && *p != '_' && *p != '-') {
	    syslog(LOG_ERR,
		"%s:%d: the book name contains invalid character: %s",
		filename, lineno, arg);
	    return -1;
	}
    }

    /*
     * Check for uniqness of book names.
     */
    if (find_book(arg) != NULL) {
	syslog(LOG_ERR, "%s:%d: the book name redefined: %s", filename,
	    lineno, arg);
	return -1;
    }

    /*
     * Set the book name.
     */
    strcpy(current_book->name, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set book-name: %s", filename,
	lineno, current_book->name);

    return 0;
}


/*
 * `title' directive (in book structure).
 */
static int
conf_book_title(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    /*
     * Check for the length of the title.
     */
    if (MAXLEN_BOOKTITLE < strlen(arg)) {
	strncpy(current_book->title, arg, MAXLEN_BOOKTITLE);
	*(current_book->title + MAXLEN_BOOKTITLE) = '\0';
	syslog(LOG_ERR, "%s:%d: too long book title: %s...", filename, lineno,
	    current_book->title);
	return -1;
    }

    /*
     * Set the title.
     */
    strcpy(current_book->title, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set book-title: %s", filename,
	lineno, current_book->title);

    return 0;
}    


/*
 * `path' directive (in book structure).
 */
static int
conf_book_path(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_FILENAME + 1];
    size_t len;

    /*
     * Check for the length of the path.
     */
    len = strlen(arg);
    if (MAXLEN_FILENAME < len) {
	strncpy(tmpbuf, arg, MAXLEN_FILENAME);
	*(tmpbuf + MAXLEN_FILENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long path: %s...", filename, lineno, 
	    tmpbuf);
	return -1;
    }

    /*
     * The path must be an absolute path.
     */
    if (*arg != '/') {
	syslog(LOG_ERR, "%s:%d: not an absolute path: %s", filename, lineno,
	    arg);
	return -1;
    }

    /*
     * Bind the book.
     */
    if (server_mode != SERVER_MODE_UTILITY
	&& eb_bind(&current_book->book, arg) < 0) {
	syslog(LOG_ERR, "%s:%d: warning: the book is not available, %s",
	    filename, lineno, eb_error_message());
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set book-path: %s", filename,
	lineno, arg);

    return 0;
}


/*
 * `appendix-path' directive.
 */
static int
conf_appendix_path(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_FILENAME + 1];
    size_t len;

    /*
     * Check for the length of the path.
     */
    len = strlen(arg);
    if (MAXLEN_FILENAME < len) {
	strncpy(tmpbuf, arg, MAXLEN_FILENAME);
	*(tmpbuf + MAXLEN_FILENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long appendix path: %s...", filename,
	    lineno, tmpbuf);
	return -1;
    }

    /*
     * The path must be an absolute path.
     */
    if (*arg != '/') {
	syslog(LOG_ERR, "%s:%d: not an absolute path: %s", filename, lineno,
	    arg);
	return -1;
    }

    /*
     * Bind the appendix.
     */
    if (server_mode != SERVER_MODE_UTILITY
	&& eb_bind_appendix(&current_book->appendix, arg) < 0) {
	syslog(LOG_ERR, "%s:%d: warning: the appendix is not available, %s",
	    filename, lineno, eb_error_message());
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set appendix-path: %s", filename,
	lineno, arg);

    return 0;
}


/*
 * `max-clients' directive (in book structure).
 */
static int
conf_book_max_clients(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];
    
    /*
     * An argument represents a remote hostname or address.
     */
    if (parse_integer(arg, &current_book->max_clients) < 0) {
	strncpy(tmpbuf, arg, MAXLEN_STRING);
	*(tmpbuf + MAXLEN_STRING) = '\0';
	syslog(LOG_ERR, "%s:%d: not an integer: %s", filename, lineno,
	    tmpbuf);
	return -1;
    }

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set book-max-clients: %d",
	filename, lineno, current_book->max_clients);

    return 0;
}


/*
 * `hosts' directive (in book structure).
 */
static int
conf_book_hosts(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    char tmpbuf[MAXLEN_STRING + 1];

    /*
     * Add `arg' to the permission list.
     */
    if (add_permission(&current_book->permissions, arg) < 0) {
	syslog(LOG_ERR, "%s:%d: memory exhausted", filename, lineno);
	return -1;
    }

    /*
     * Debugging information.
     */
    strncpy(tmpbuf, arg, MAXLEN_STRING);
    *(tmpbuf + MAXLEN_STRING) = '\0';
    syslog(LOG_DEBUG, "%s:%d: debug: add book-hosts: %s", filename,
	lineno, tmpbuf);

    return 0;
}


/*
 * `alias-eiwa' directive (in book structure).
 */
static int
conf_book_alias_eiwa(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    const char *p;

    /*
     * Check for the length of the subbook name.
     */
    if (EB_MAXLEN_BASENAME < strlen(arg)) {
	strncpy(alias_eiwa, arg, EB_MAXLEN_BASENAME);
	*(alias_eiwa + EB_MAXLEN_BASENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long aliased name: %s...", filename,
	    lineno, alias_eiwa);
	return -1;
    }

    /*
     * The subbook name must consist of upper letters, digit, '_' and '-'.
     */
    for (p = arg; *p != '\0'; p++) {
	if (!isupper(*p) && *p != '_' && *p != '-') {
	    syslog(LOG_ERR, "%s:%d: the name contains invalid character: %s",
		filename, lineno, arg);
	    return -1;
	}
    }

    /*
     * Check for redefinition.
     */
    if (alias_eiwa[0] != '\0') {
	syslog(LOG_ERR, "%s:%d: redifined: alias-eiwa", filename, lineno);
	return -1;
    }

    /*
     * Set the alias `eiwa'.
     */
    strcpy(alias_eiwa, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set alias-eiwa: %s", filename,
	lineno, alias_eiwa);

    return 0;
}


/*
 * `alias-waei' directive (in book structure).
 */
static int
conf_book_alias_waei(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    const char *p;

    /*
     * Check for the length of the subbook name.
     */
    if (EB_MAXLEN_BASENAME < strlen(arg)) {
	strncpy(alias_waei, arg, EB_MAXLEN_BASENAME);
	*(alias_waei + EB_MAXLEN_BASENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long aliased name: %s", filename, lineno,
	    alias_waei);
	return -1;
    }

    /*
     * The subbook name must consist of upper letters, digit, '_' and '-'.
     */
    for (p = arg; *p != '\0'; p++) {
	if (!isupper(*p) && *p != '_' && *p != '-') {
	    syslog(LOG_ERR, "%s:%d: the name contains invalid character: %s",
		filename, lineno, arg);
	    return -1;
	}
    }

    /*
     * Set the alias `waei'.
     */
    if (alias_waei[0] != '\0') {
	syslog(LOG_ERR, "%s:%d: redefined: alias-waei", filename, lineno);
	return -1;
    }

    /*
     * Set the alias `waei'.
     */
    strcpy(alias_waei, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set alias-waei: %s", filename,
	lineno, alias_waei);

    return 0;
}


/*
 * `alias-kojien' directive (in book structure).
 */
static int
conf_book_alias_kojien(blockname, dirname, arg, filename, lineno)
    const char *blockname;
    const char *dirname;
    const char *arg;
    const char *filename;
    int lineno;
{
    const char *p;

    /*
     * Check for the length of the subbook name.
     */
    if (EB_MAXLEN_BASENAME < strlen(arg)) {
	strncpy(alias_kojien, arg, EB_MAXLEN_BASENAME);
	*(alias_kojien + EB_MAXLEN_BASENAME) = '\0';
	syslog(LOG_ERR, "%s:%d: too long aliased name: %s", filename, lineno,
	    alias_kojien);
	return -1;
    }

    /*
     * The subbook name must consist of upper letters, digit, '_' and '-'.
     */
    for (p = arg; *p != '\0'; p++) {
	if (!isupper(*p) && *p != '_' && *p != '-') {
	    syslog(LOG_ERR, "%s:%d: the name contains invalid character: %s",
		filename, lineno, arg);
	    return -1;
	}
    }

    if (alias_kojien[0] != '\0') {
	syslog(LOG_ERR, "%s:%d: redifined: alias-kojien", filename, lineno);
	return -1;
    }

    /*
     * Set the alias `kojien'.
     */
    strcpy(alias_kojien, arg);

    /*
     * Debugging information.
     */
    syslog(LOG_DEBUG, "%s:%d: debug: set alias-kojien: %s", filename,
	lineno, alias_kojien);

    return 0;
}


